#include <string>
#include <iostream>
#include <memory>
#include <algorithm>
#include <time.h>
#include "FDS/rt_api.h"
#include "FDS/rt_fields.h"
#include "FDS/Level2Feature.h"
#include <fire/Logger.h>

using namespace std;
using namespace FDS;

class RTClient
{
  public:

  RTClient(const string &connection);
  RTClient(const string& user, const string& host, const string& path);
  int start();
  int stop();
  int make_request(const string &symbol);
  int main_loop();

  static void handle_message(int tag, const RT_Message *msg,
                              const RT_Record   *rec, void *closure);

  static void handle_connection_change(bool is_connected, const RT_Message *msg,
                                       void *closure);
  // our state is pretty simple
  string connection_string;
  bool   exit_main_loop;
  int    num_messages;
  // OTP variables
  string user_serial;
  string otp_path;
  string host;
  bool is_otp;
  clock_t start_time;
  int duration;
  string log_file;

  set<int> tag_set;

  // snapshot ignore variable
  bool ignore_snapshot;
  uint64_t snapshot_bytes;
  int snapshot_msg_count;

  private:
  RTClient(const RTClient &not_copyable);
  RTClient& operator=(const RTClient &not_copyable);
};

RTClient::RTClient(const string &connection) : connection_string(connection), log_file(""), snapshot_bytes(0), snapshot_msg_count(0), ignore_snapshot(false)
{
  is_otp = false;
  exit_main_loop = false;
  num_messages = 0;
  RT_Consumer::get_default().set_connection_ctrl_cb(&RTClient::handle_connection_change, this);
}

RTClient::RTClient(const string& user, const string& host_name, const string& path)
                    : user_serial(user), otp_path(path) , host(host_name),
                      log_file(""), snapshot_bytes(0), snapshot_msg_count(0) {
  is_otp = true;
  exit_main_loop = false;
  num_messages = 0;
  RT_Consumer::get_default().set_connection_ctrl_cb(&RTClient::handle_connection_change, this);
}

int RTClient::start()
{

  string fields_file = "../../../etc/rt_fields.xml";

  if(log_file != "") FDF::log_open(log_file.c_str(), false);

  // Open our map file
  RT_FieldMap *map = RT_FieldMap::create(fields_file.c_str());
  if (!map)
    RT_LOG_WARN("Unable to open the field map file: " << fields_file);
  rt_errno err;
  if (!is_otp) {
    // Set the connection info (user:pass@host:port), from the connection string
    err = RT_Consumer::get_default().set_connection_info(connection_string.c_str());
    if (err) {
      RT_LOG_ERROR("Unable to get connection info from registry e:" << err);
      return err;
    }
  }
  else {
    size_t sep_loc = otp_path.find_last_of("/\\");
    string path = "";
    string device_id = "";

    if (sep_loc != std::string::npos)
    {
      path = otp_path.substr(0, sep_loc);
      device_id = otp_path.substr(sep_loc + 1);
    }
    else
    {
      device_id = otp_path;
    }

    // Removing .data extension from the key-counter data file
    size_t ext_loc = device_id.find(".data\0");
    if (ext_loc != std::string::npos)
    {
      device_id = device_id.substr(0, ext_loc);
    }

    err = RT_Consumer::get_default().set_connection_info(host.c_str(),
      user_serial.c_str(),
      device_id.c_str(),
      NULL,
      NULL,
      path.c_str(),
      false);
    if (err) {
      RT_LOG_ERROR("Unable to connect using OTP e:" << err);
      return err;
    }
  }

  // Connect to the data server
  err = RT_Consumer::get_default().connect();
  if (err) {
    RT_LOG_ERROR("Connection to RT_Server[" << RT_Consumer::get_default().get_connection_info()
                 << "] failed err:" << err);
    return err;
  }
  return 0;
}

int RTClient::stop()
{
  // cancel the remaining tags before disconnecting
  for (auto tag : tag_set) {
    //RT_LOG_INFO("cancelling tag " << tag);
    RT_Consumer::get_default().cancel(tag);
  }
  return RT_Consumer::get_default().disconnect();
}

// Static method, just log the message
void RTClient::handle_connection_change(bool is_connected, const RT_Message *msg,
                                      void *closure)
{
  RT_LOG_INFO("Control Message: " << *msg);
  if (!is_connected) RT_LOG_ERROR("server disconnected!! :(");
}

// Static method, just print the message to our log file
void RTClient::handle_message(int tag, const RT_Message *msg, const RT_Record *rec, void *closure)
{
  RTClient *clnt = static_cast<RTClient *>(closure);
  RT_Field fld = msg->get(FIDS::MSG_TYPE);
  if (!fld.empty() && fld.to_string() == "S" && clnt->ignore_snapshot) {
    // snapshot msg so save the bytes store which can be ignored at last
    clnt->snapshot_bytes = RT_Consumer::get_default().get_bytes_read_count();
    clnt->snapshot_msg_count = RT_Consumer::get_default().msg_count();
  }

  if ((clock() - clnt->start_time) / CLOCKS_PER_SEC >= clnt->duration) {
    // time reached so stop all dispatching
    RT_Consumer::get_default().cancel(tag);
    clnt->tag_set.erase(tag);
    clnt->exit_main_loop = true;
    //RT_LOG_INFO("Time Limit reached. Cancelling tag " << tag);
  }
  if (msg->is_closed()) {
    RT_LOG_ERROR("Server closed stream");
    RT_Consumer::get_default().cancel(tag);
    clnt->tag_set.erase(tag);
    if (clnt->tag_set.size() == 0) // all clients are canceled, so stop the dispatch loop
      clnt->exit_main_loop = true;
    if (msg->is_error()) RT_LOG_ERROR(msg->get_error_description(NULL) << " for tag " << tag);
  }

  if (msg->is_stale())
    RT_LOG_INFO("STALE!!!!! reason: " << msg->get_stale_reason(NULL));
}

// Make a real-time request
int RTClient::make_request(const string &symbol)
{
  int tag;
  RT_Request req("FDS1", (symbol).c_str());
  int err = RT_Consumer::get_default().request(req, &RTClient::handle_message, this, &tag);

  if (err)
    RT_LOG_WARN(req << " err:" << err);
  else {
    tag_set.insert(tag);
    RT_LOG_INFO(req << " tag:" << tag);
  }
  return err;
}

int RTClient::main_loop()
{
  int err = 0;
  while (!exit_main_loop) {
    if ((clock() - start_time) / CLOCKS_PER_SEC >= duration) {
      RT_LOG_INFO("Terminating dispatch loop because of timeout");
      break;
    }
    err = RT_Consumer::get_default().dispatch(1000);
    if (err) {
      RT_LOG_ERROR("Error in dispatch e:" << err);
      break;
    }
  }
  return err;
}

// Helper class for parsing the command line arguments
class InputParser {
public:
  InputParser(int &argc, char **argv) {
    for (int i = 1; i < argc; ++i)
      this->tokens.push_back(string(argv[i]));
  }
  vector<string> get_multiple_items(const string &option) const {
    vector<string>::const_iterator itr;
    vector<string> symbols;
    itr = find(this->tokens.begin(), this->tokens.end(), option);
    ++itr;
    while (itr != this->tokens.end()) {
      if (itr->at(0) == '-') break; // dont add if it is a command
      symbols.push_back(*itr);
      ++itr;
    }
    return symbols;
  }

  const string get_single_item(const string &option) const {
    vector<string>::const_iterator itr;
    itr = find(this->tokens.begin(), this->tokens.end(), option);
    if (itr != this->tokens.end() && ++itr != this->tokens.end()) {
      return *itr;
    }
    return string("");
  }

  bool is_option_exists(const string &option) const {
    return find(this->tokens.begin(), this->tokens.end(), option)
      != this->tokens.end();
  }
private:
  vector <string> tokens;
};

//
// example usage:
// .\BandwidthEstimate -c user:passwd@api-stage.df.factset.com -s IBM-USA
// .\BandwidthEstimate -u user-serial -p ./AAAA.data -h api-stage.df.factset.com -s IBM-USA
//
int main(int argc, char **argv)
{
  if (argc <= 4) {
    RT_LOG_RAW("Basic authentication \n Usage: " << argv[0] << "-c {user-serial:passwd@host} -s {symbol} {symbol} .. {symbol}\n [-t time_limit] [-o output_file] [-x] [-l2]\n");
    RT_LOG_RAW("Example: BandwidthEstimate -c user:passwd@api-stage.df.factset.com -s IBM-USA\n");
    RT_LOG_RAW("OTP authentication \n Usage: " << argv[0] << " -u {user-serial} -p {full-path-to-OTP-file} -h {host} -s {symbol} {symbol} .. {symbol}\n");
    RT_LOG_RAW("Example: BandwidthEstimate -u user-serial -p ./AAAA.data -h api-stage.df.factset.com -s IBM-USA\n");
    RT_LOG_RAW("--symbols,s symbol1 symbol2 (symbols seperated by space subscribed to FDS1. alternative to --symbols-file)\n");
    RT_LOG_RAW("--symbols-file,f  file_name_containing_symbols_seperated_by_newline_or_comma\n");
    RT_LOG_RAW("--connection-string,c  user-serial:passwd@host (either this is required or -p with OTP authentication)\n");
    RT_LOG_RAW("--time-limit,t  time_limit_in_seconds (defaults to 120 if not specified)\n");
    RT_LOG_RAW("--output-file,o   file_where_output_should_be_written (defaults to stdout)\n");
    RT_LOG_RAW("--ignore-snapshot,x  (option to ignore snapshot in bandwidth calculation, off by default)\n");
    RT_LOG_RAW("-l2 option is used to enable level 2 data\n");
    return 3;
  }

  InputParser input(argc, argv);
  auto_ptr<RTClient> clnt;
  if (input.is_option_exists("-p")) { // otp authentication
    string user_serial = input.get_single_item("-u");
    string otp_path = input.get_single_item("-p");
    string host = input.get_single_item("-h");
    if (user_serial.empty() || otp_path.empty() || host.empty()) {
      RT_LOG_RAW("For OTP please provide all three options ( -p -u -h )\n");
      return 3;
    }
    clnt = auto_ptr<RTClient>(new RTClient(user_serial, host, otp_path));
  }
  else { // basic authentication
    string connection_string = (input.is_option_exists("-c")) ? input.get_single_item("-c") : input.get_single_item("--connection-string");
    if (connection_string.empty()) {
      RT_LOG_RAW("Connection string is empty. Please add a valid connection string with option -c\n");
      return 3;
    }
    clnt = auto_ptr<RTClient>(new RTClient(connection_string));
  }
  if (input.is_option_exists("-t") || input.is_option_exists("--time-limit")) {
    clnt->duration = (input.is_option_exists("-t")) ?stoi(input.get_single_item("-t")) : stoi(input.get_single_item("--time-limit"));
  } else {
    // default is 120 seconds
    clnt->duration = 120;
  }

  vector<string> symbols;
  if (input.is_option_exists("-f") || input.is_option_exists("--symbols-file")) {
    string input_file = (input.is_option_exists("-f")) ?input.get_single_item("-f") : input.get_single_item("--symbols-file");
    freopen(input_file.c_str(), "r", stdin);
    string line;
    // read every line from the stream
    while (getline(std::cin, line))
    {
      istringstream csvStream(line);
      string symbol;
      while (getline(csvStream, symbol, ','))
      {
        symbol.erase(remove(symbol.begin(), symbol.end(), ' '), symbol.end()); // remove space
        symbol.erase(remove(symbol.begin(), symbol.end(), '\t'), symbol.end()); // remove tabs if present
        if (symbol.size() > 0)
          symbols.push_back(symbol);
      }
    }
  }
  else if (input.is_option_exists("-s") || input.is_option_exists("--symbols")) {
    symbols = (input.is_option_exists("-s")) ? input.get_multiple_items("-s") : input.get_multiple_items("--symbols");
  }
  else {
    RT_LOG_RAW("Symbol list is empty. Please add a valid symbols using either -s option or -f option\n");
    return 3;
  }

  if (input.is_option_exists("-x") || input.is_option_exists("--ignore-snapshot")) {
    clnt->ignore_snapshot = true;
  }
  if (input.is_option_exists("-o") || input.is_option_exists("--output-file")) {
    clnt->log_file = (input.is_option_exists("-o")) ? input.get_single_item("-o") : (input.get_single_item("--output-file"));
  }
  RT_LOG_INFO("INFO: about to connect");
  int err = clnt->start();
  if (err) {
    RT_LOG_ERROR("Error in connect e: " << err);
    return err;
  }
  RT_LOG_INFO("connected, make request");

  bool    is_level2 = input.is_option_exists("-l2");
  if(is_level2) {
    if (!RT_Consumer::get_default().is_active(Level2Feature())) {
      bool res = RT_Consumer::get_default().enable(Level2Feature());
      if (!res) {
        RT_LOG_ERROR("Error to enable Level 2 feature.");
        is_level2 = false;
      }
    }
  }
  while (!symbols.empty()) {
    string cur_symbol = symbols.back();
    if(is_level2 && cur_symbol.find(":L2") == string::npos)
      cur_symbol += ":L2"; // add only if ":L2" is not already present
    clnt->make_request(cur_symbol);
    symbols.pop_back();
  }

  // Enter main loop
  clnt->start_time = clock();
  int rtn = clnt->main_loop();

  // disconnect
  RT_LOG_INFO("main loop rtn'd: " << rtn);

  // Calculating the bandwidth
  uint64_t bytes_read = RT_Consumer::get_default().get_bytes_read_count();
  uint64_t msg_count = RT_Consumer::get_default().msg_count();

  RT_LOG_INFO("***************************************************************");
  if (clnt->ignore_snapshot) {
    RT_LOG_INFO("Ignoring snapshot for calculating bandwidth");
    bytes_read -= clnt->snapshot_bytes;
    msg_count -= clnt->snapshot_msg_count;
  }
  double avg_bandwidth_per_msg = bytes_read / (double)msg_count;
  RT_LOG_INFO("Bandwidth calculated for " << clnt->duration << " second(s).");
  RT_LOG_INFO("Number of messages\t: " << msg_count);
  RT_LOG_INFO("Average bandwidth\t: " << avg_bandwidth_per_msg << " Byte(s)/message");
  RT_LOG_INFO("Total bandwidth\t: " << bytes_read << " Bytes.");
  RT_LOG_INFO("***************************************************************");

  clnt->stop();
}

