#include <iostream>
#include <string>
#include <cstring>
#include <cstdlib>
#include <fstream>
#include <errno.h>
#include <vector>
#include <assert.h>
#include <time.h>
#include "FDS/rt_api.h"
#include "FDS/rt_fields.h"

using namespace std;
using namespace FDS;

// Uppercase is always tougher to type :>
#define log_info  RT_LOG_INFO
#define log_error RT_LOG_ERROR
#define log_warn  RT_LOG_WARN

// Configuration filename.  Can be changed via the first argument to main()
#define CONFIGURATION_FILE "csv_file_cache.cfg"

struct Subscription {
  Subscription(const string &sym) : req_sym(sym), tag(-1) {}

  string req_sym;    // Our requested symbol (from the symbol file)
  int    tag;        // The API subscription tag
};

// Just use a simple vector of records to store our subscriptions
vector<Subscription> subscriptions;

//
// Function to read in the symbol file
//
int read_symbol_file(const char *file)
{
  ifstream infile(file);
  if (!infile) {
    log_error("Unable to open symbol file: " << file << " err:" << errno);
    return (errno) ? errno : -1;
  }

  char buff[513];
  buff[512] = 0;

  while (infile.getline(buff, sizeof(buff)-1)) {
    char *pp = buff;

    // Get rid of C++ style comments.  Note: If we do this we may have to convert
    char *comment = strstr(pp, "//");
    if (comment) *comment = '\0';

    // Also skip shell style comments (first char = '#') and blank lines
    if (*pp == 0 || *pp == '#')
      continue;

    subscriptions.push_back(Subscription(buff));
    //log_info("got input symbol = " << symbol);
  }

  return 0;
}

size_t split(vector<std::string> *dest, const char *src, char delim)
{
  const char *delim_p;
  size_t num_items = 0;
  std::string item;
  assert(dest);
  while ((delim_p = strchr(src, delim))) {
    item.assign(src, delim_p);
    dest->push_back(item);
    ++num_items;
    src = delim_p + 1;
  }
  item = src;
  dest->push_back(item);
  ++num_items;
  return num_items;
}

//
// Function to print the output file
//
void print_output_file(const string &file_name, const string &fields,
                       const vector<int> &fids)
{
  rt_errno rterr;
  const RT_Record *rec;
  size_t nsubs = subscriptions.size();

  log_info("writing csv file for " << (unsigned int)nsubs << " subscriptions");

  ofstream ofile;
  string tmp_file(file_name);
  tmp_file += ".tmp";

  // Open a temp file (to do the writing)
  ofile.open(tmp_file.c_str(), ios::out | ios::trunc);
  if (!ofile.is_open()) {
    log_error("Unable to open outfile: " << tmp_file << " e:" << errno);
    return;
  }

  // Write header row
  ofile << "REQ_SYM,KEY," << fields << '\n';

  // Write the field data
  for (unsigned int i = 0; i < nsubs; ++i, ofile << '\n') {
    ofile << subscriptions[i].req_sym;

    rterr = FDF::get_record(subscriptions[i].tag, &rec);
    if (rterr) {
      ofile << ",{Waiting for Response}";  
      for (unsigned int j=0; j < fids.size(); ++j)
        ofile << ",";
      continue;
    }

    ofile << ',' << rec->get_key().to_string();
    for (unsigned int j=0; j < fids.size(); ++j)
      ofile << ',' << rec->get(fids[j]).to_string();
  }

  ofile.flush();
  ofile.close();

  // Now move the temp file to the real file name
  remove(file_name.c_str());
  rename(tmp_file.c_str(), file_name.c_str());
  log_info("done");
}

int main(int argc, char* argv[])
{
  string config_file = (argv[1]) ? (argv[1]) : CONFIGURATION_FILE;

  // open our log file
  // FDF::log_open("csv_file_cache.log");

  log_info("CSV File Cache started.  Opening config file " << config_file);

  rt_errno rterr = FDF::load_properties(config_file.c_str());
  if (rterr) {
    log_error("Couldn't load properties: " << config_file
              << " e:" << rterr);
    return rterr;
  }

  // Generic buffer to get our configuration values
  char prop_val[FDF::MAX_PROPERTY_SIZE];

  // Get the field map file
  RT_FieldMap *map;
  FDF::get_property("FIELD_MAP_FILE", prop_val);
  map = RT_FieldMap::create(prop_val);
  if (!map) {
    log_error("Unable to load field map file: " << prop_val);
    return -1;
  }

  // Get the fields of interest, split them and convert them to integer fids
  FDF::get_property("FIELDS", prop_val);
  string fields(prop_val);
  vector<string> fields_v;
  vector<int>    fids_v;
  split(&fields_v, prop_val, ',');
  for (unsigned int i=0; i < fields_v.size(); ++i)
    fids_v.push_back(map->get_id(fields_v[i].c_str()));

  // Get our symbol file (defaults to symbols.dat)
  // Read in the symbols.  Exit program if we can't read in our tickers.
  FDF::get_property("SYMBOL_FILE", prop_val, "symbols.dat");
  if (read_symbol_file(prop_val)) return -1;

  // Get our output timer (defaults to 5 minutes) and file
  FDF::get_property("OUTPUT_TIMER_SEC", prop_val, "300");
  unsigned int output_timer_secs = (unsigned int)atoi(prop_val);
  if (!output_timer_secs) {
    log_warn("Invalid output timer, must be non-zero val= " << output_timer_secs);
    output_timer_secs = 300;
  }

  // Output file defaults to "symbols.csv"
  FDF::get_property("OUTPUT_FILE", prop_val, "symbols.csv");
  string out_file(prop_val);
 
  // Get our data service (defaults to "FDS1")
  FDF::get_property("DATA_SERVICE", prop_val, "FDS1");
  string data_service(prop_val);

  // Get OTP details
  FDF::get_property("OTP_FULL_PATH", prop_val, "");
  string otp_full_path(prop_val);

  if(otp_full_path.empty()) {
    // Passing NULL into set_connection_info will cause it to look for the
    // RT_CONNECTION property using FDF::get_property() 
    FDF::set_connection_info(NULL);
  } else {
    // Get host details
    FDF::get_property("RT_HOST", prop_val, "");
    string host(prop_val);

    // Get user details
    FDF::get_property("RT_USER", prop_val, "");
    string user_serial(prop_val);

    string path = "";
    string device_id = "";
    size_t sep_loc = otp_full_path.find_last_of("\\");
    if(sep_loc!=string::npos) {
      path = otp_full_path.substr(0, sep_loc);
      device_id = otp_full_path.substr(sep_loc+1);
    } else {
      device_id = otp_full_path;
    }
    size_t dot_loc = device_id.find(".");
    if(dot_loc!=string::npos) {
      device_id = device_id.substr(0, dot_loc);
    }
    log_info("OTP connection details host:" << host << " user:" << user_serial << " device_id:" << device_id << " in path:" << path);
    rterr = FDF::set_connection_info(host.c_str(), user_serial.c_str(), device_id.c_str(), NULL, NULL, path.c_str(), false);
    if(rterr) {
      log_error("Error in connecting through otp e:" << rterr);
      return rterr;
    }
  }

  // Connect to the server
  rterr = FDF::connect();
  if (rterr) {
    log_error("Unable to connect " << rterr);
    return -1;
  }

  // Make real-time requests for all symbols in our price_file
  for (unsigned int i=0; i < subscriptions.size(); ++i) {
    int tag;

    // Make a request.
    // Note: we can pass in a NULL for the callback. We just want the API
    // to subscribe, and keep the record up to date.  We are not interested
    // in the updates.
    RT_Request req(data_service.c_str(), subscriptions[i].req_sym.c_str());
    rterr = FDF::request(req, NULL, NULL, &tag);

    log_info("requested " << subscriptions[i].req_sym << " tag=" << tag);
    if (rterr) log_warn("Request is queued e: " << rterr);

    // Set the state that we know of (like tag, and requested key)
    subscriptions[i].tag  = tag;

    // When requesting many tickers (like > 5000), it is better to dispatch
    // messages while requesting.  This way the system is not queuing up
    // messages internally (since there is an API thread reading the TCP
    // socket).  Note:  This is optional, and can be commented out if desired.
    rterr = FDF::dispatch(0);
  }

  // Initialize our timer variables
  time_t curr_time, prev_time;
  curr_time = prev_time = time(0);
  unsigned int elapsed_time = 0; 

  // Our main loop
  while (1) {
    rterr = FDF::dispatch((output_timer_secs - elapsed_time) * 1000);
    if (rterr) {
      log_warn("dispatch err: " << rterr);
    }
    curr_time = time(0);
    elapsed_time = elapsed_time + (unsigned int)(curr_time - prev_time);

    if (elapsed_time >= output_timer_secs) {
      // time to write our output file, write it and reset elapsed_time
      print_output_file(out_file, fields, fids_v);
      elapsed_time = 0;
    }

    prev_time = curr_time;
  }
}


